/* 
 *     PROJECT          : emailEJB 
 *
 *     COPYRIGHT        : Copyright (C) 1999-2003 Tim Stephenson
 *
 *     contact me at gpl@enableit.org.
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 
package org.enableit.tools.email.ejb;

// Java Imports
import java.io.File ;
import java.io.IOException ;
import java.net.URL ;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;

import javax.activation.DataHandler;
import javax.activation.FileDataSource ;
import javax.activation.URLDataSource ;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import javax.mail.BodyPart;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Part;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;

// required for SSL URL connections
import java.security.Security ;

// HTTP Client imports
import org.apache.commons.httpclient.HttpClient ;
import org.apache.commons.httpclient.methods.GetMethod ;

// Log4J Imports
import org.apache.log4j.Logger;

import org.enableit.tools.email.interfaces.Attachment ;
import org.enableit.tools.email.interfaces.EmailConfigBean ;

/**
 * An EJB for sending email. 
 *
 * @see <a href="http://darrt.sourceforge.net/emailEJB/index.html">
 *      Main documentation</a>
 * @author Tim Stephenson
 *
 * @ejb:bean type="Stateless" name="Email" jndi-name="Email"
 *      display-name="Email"
 * @ejb:transaction type="Supports"
 * @ejb:transaction-type type="Container"
 * @ejb:util generate="physical"
 * @ejb:env-entry name="HOST" type="java.lang.String" value="mail.mydomain.com"
 * @ejb:env-entry name="MBOX" type="java.lang.String" value="null"
 * @ejb:env-entry name="PROTOCOL" type="java.lang.String" value="SMTP"
 * @ejb:env-entry name="USER" type="java.lang.String" value="username"
 * @ejb:env-entry name="PASSWD" type="java.lang.String" value="password"
 * @ejb:env-entry name="SEND" type="java.lang.String" value="send"
 * @ejb:env-entry name="MAILER" type="java.lang.String" value="msgsend"
 * @ejb:env-entry name="FROM" type="java.lang.String"
 *      value="username@mydomain.com"
 * @ejb:env-entry name="DEBUG" type="java.lang.Boolean" value="Boolean.TRUE"
 */
public class EmailBean extends Object {

// Constructors  --------------------------------------------------------------

    /**
     * Default constructor.
     * This is the only constructor allowable for an EJB
     */
    public EmailBean() {

    }

// Lifecycle Methods  ---------------------------------------------------------

    /**
     * Called by container when creating new bean instance.
     * @see <a href="http://java.sun.com/products/ejb/">EJB spec</a>
     */
    public void ejbCreate()
            throws javax.ejb.CreateException, javax.ejb.EJBException {
        initialise() ;
    }

// Business Methods  ----------------------------------------------------------

    /**
     * Initialise the bean configurations from JNDI tree.
     */
    protected void initialise() {
        // Log the entry into the method.
        logger.info("METHOD_ENTRY initialise");

        if (!initialised) {
            // Lookup the environment vars (these are mandatory)
            Context myEnv = null ;
            try {
                Security.addProvider(
                        new com.sun.net.ssl.internal.ssl.Provider());
                // TODO: Allow to be overridden by JNDI environment property
                System.setProperty("java.protocol.handler.pkgs",
                        "com.sun.net.ssl.internal.www.protocol") ;

                Context ic = new InitialContext();
                myEnv = (Context) ic.lookup("java:comp/env");
                // Get the mail server config
                protocol = (String) myEnv.lookup("PROTOCOL");
                logger.warn("Email:protocol:" + protocol) ;
                mbox    = (String) myEnv.lookup("MBOX");
                logger.warn("Email:mbox:" + mbox) ;
                send    = (String) myEnv.lookup("SEND");
                logger.warn("Email:send:" + send) ;
                host    = (String) myEnv.lookup("HOST");
                logger.warn("Email:host:" + host) ;
                user    = (String) myEnv.lookup("USER");
                logger.warn("Email:user:" + user) ;
                passwd  = (String) myEnv.lookup("PASSWD");
                logger.warn("Email:passwd:" + passwd) ;
                mailer  = (String) myEnv.lookup("MAILER");
                logger.warn("Email:mailer:" + mailer) ;
                from    = (String) myEnv.lookup("FROM");
                logger.warn("Email:from:" + from) ;
            } catch (NamingException ne) {
                logger.error(ne.getMessage()) ;
            }

            try {
                debug   = (Boolean) myEnv.lookup("DEBUG");
                logger.warn("Email:debug:" + debug.toString()) ;
            } catch (NamingException ne) {
                logger.warn("No env var 'DEBUG' found, "
                        + " mail debugging will be switched off.") ;
            }
            try {
                String tmpBlankSubject = (String)
                        myEnv.lookup("BLANK_SUBJECT");
                if (tmpBlankSubject != null && tmpBlankSubject.length() > 0) {
                    blankSubject = tmpBlankSubject ;
                }
                logger.warn("Email:blankSubject:" + blankSubject) ;
            } catch (NamingException ne) {
                logger.warn("No env var 'BLANK_SUBJECT' found.") ;
            }

            initialised = true ;
        }

        // Log the exit from the method.
        logger.info("METHOD_EXIT initialise");
    }

    /**
     * Initialise the bean configurations from params.
     *
     * <p>This method is really only intended for testing purposes,
     * though it could also be used to initialise this class as a
     * simple class, rather than an EJB.</p>
     * @param protocol The mail protocol to configure with.
     * @param mbox Not used since no mailbox is ever read.
     * @param send JavaMail property 'send'
     * @param host The outgoing mail host.
     * @param user User to authenticate with on the outgoing mail server.
     * @param passwd Password (plain text) to authenticate with on the
     *      outgoing mail server.
     * @param mailer JavaMail property 'mailer'
     * @param from Default for the mail 'from' field if not provided on send.
     * @param debug Whether to use a debug mail session.
     * @param blankSubject The subject to use when the caller of sendEmail
     *      does not provide one (may be blank but may <em>not</em> be null).
     */
    protected void initialise(String protocol, String mbox, String send,
            String host, String user, String passwd, String mailer,
            String from, Boolean debug, String blankSubject) {
        // Log the entry into the method.
        logger.info("METHOD_ENTRY initialise");

        if (!initialised) {
            Security.addProvider(new com.sun.net.ssl.internal.ssl.Provider());
            // TODO: Allow to be overridden by JNDI environment property
            System.setProperty("java.protocol.handler.pkgs",
                    "com.sun.net.ssl.internal.www.protocol") ;

            // Get the mail server config
            this.protocol = protocol;
            logger.warn("Email:protocol:" + protocol) ;
            this.mbox = mbox;
            logger.warn("Email:mbox:" + mbox) ;
            this.send = send;
            logger.warn("Email:send:" + send) ;
            this.host = host;
            logger.warn("Email:host:" + host) ;
            this.user = user;
            logger.warn("Email:user:" + user) ;
            this.passwd = passwd;
            logger.warn("Email:passwd:" + passwd) ;
            this.mailer = mailer;
            logger.warn("Email:mailer:" + mailer) ;
            this.from = from;
            logger.warn("Email:from:" + from) ;

            this.debug = debug;
            logger.warn("Email:debug:" + debug.toString()) ;

            String tmpBlankSubject = blankSubject;
            if (tmpBlankSubject != null && tmpBlankSubject.length() > 0) {
                this.blankSubject = tmpBlankSubject ;
            }
            logger.warn("Email:blankSubject:" + blankSubject) ;

            initialised = true ;
        }

        // Log the exit from the method.
        logger.info("METHOD_EXIT initialise");
    }

    /**
     * Sends an email message using the Mail server configured in the JNDI
     * environment.
     *
     * @param to Target email address.
     * @param from Message 'from' address. If supplied this overrides the
     *      default in the JNDI context.
     * @param subject Subject for the mail message.
     * @param message Body text for the mail message.
     * @throws java.rmi.RemoteException System exception, notably
     *      misconfiguration of mail server.
     *
     * @ejb:interface-method view-type="remote"
     */
    public void sendEmail(String to, String from,
        String subject, String message)
    throws  java.rmi.RemoteException {
        logger.info("METHOD_ENTRY sendEmail");

        EmailConfigBean bean =
                new EmailConfigBean(to, from, subject, message) ;

        this.sendEmail(bean) ;

        logger.info("METHOD_EXIT sendEmail");
    }

    /**
     * Sends an email message using the Mail server configured in the JNDI
     * environment and creates an attachment too.
     *
     * @param to Target email address.
     * @param from Message 'from' address. If supplied this overrides the
     *      default in the JNDI context.
     * @param subject Subject for the mail message.
     * @param message Body text for the mail message.
     * @param attachmentText Text to create as a text/plain attachment.
     * @throws java.rmi.RemoteException System exception, notably
     *      misconfiguration of mail server.
     *
     * @ejb:interface-method view-type="remote"
     */
    public void sendEmail(String to, String from,
            String subject, String message, String attachmentText)
    throws  java.rmi.RemoteException {
        logger.info("METHOD_ENTRY sendEmail");

        EmailConfigBean bean =
                new EmailConfigBean(to, from, subject, message) ;
        Attachment attachment = null ;
        if ((attachmentText.indexOf("<html>") != -1)
                || (attachmentText.indexOf("<HTML>") != -1)) {
            attachment = new Attachment(attachmentText, MT_HTML,
                    "attachment.html") ;
        } else {
            attachment = new Attachment(attachmentText, MT_PLAIN_TEXT,
                    "attachment.txt") ;
        }
        bean.addAttachment(attachment) ;

        this.sendEmail(bean) ;

        logger.info("METHOD_EXIT sendEmail");
    }

    /**
     * Sends an email made up of a message and an attachment.
     *
     * Converts a URL reference to an InputStream.
     * The URL must be accessible using the http(s) protocol.
     *
     * @param to Target email address.
     * @param from Message 'from' address. If supplied this overrides the
     *      default in the JNDI context.
     * @param subject Subject for the mail message.
     * @param message Body text for the mail message.
     * @param attachmentUrl URL to fetch and create as an attachment.
     * @throws java.rmi.RemoteException System exception, notably
     *      misconfiguration of mail server.
     *
     * @ejb:interface-method view-type="remote"
     */
    public void sendEmail(String to, String from,
            String subject, String message, URL attachmentUrl)
    throws java.rmi.RemoteException {
        logger.info("METHOD_ENTRY sendEmail, url="
                + attachmentUrl.toExternalForm());

        try {
            EmailConfigBean bean =
                    new EmailConfigBean(to, from, subject, message) ;
            Attachment attachment = new Attachment(attachmentUrl) ;
            bean.addAttachment(attachment) ;

            this.sendEmail(bean) ;

        } catch (java.io.IOException e) {
            logger.error("MessagingException: " + e.getMessage(), e) ;
            throw new java.rmi.RemoteException(e.toString()) ;
        }

        logger.info("METHOD_EXIT sendEmail");
    }

    /**
     * Sends an email made up of a message and an attachment.
     *
     * Converts a URL reference to an InputStream.
     * The URL must be accessible using the http(s) protocol.
     *
     * @param bean An <code>EmailConfigBean</code> containing all the details
     *      for the message.
     * @throws java.rmi.RemoteException System exception, notably
     *      misconfiguration of mail server.
     *
     * @ejb:interface-method view-type="remote"
     */
    public void sendEmail(EmailConfigBean bean)
    throws java.rmi.RemoteException {
        logger.info("METHOD_ENTRY sendEmail(EmailConfigBean)");

        // Get the basic message
        Message msg = getBaseMessage(bean.getTo(), bean.getFrom(),
                bean.getSubject(), bean.getMessage());

        // Construct holder for the message
        MimeMultipart holder = new MimeMultipart();
        // Construct the main body of message
        BodyPart body = new MimeBodyPart() ;

        try {
            // sort the body
            body.setDisposition(Part.INLINE);
            body.setContent(bean.getMessage(), MT_PLAIN_TEXT) ;
            holder.addBodyPart(body);

            // sort the attachment(s)
            try {
                for (Iterator i = bean.getAttachmentList().iterator() ;
                        i.hasNext() ;) {
                    BodyPart attachment = new MimeBodyPart() ;
                    attachment.setDisposition(Part.ATTACHMENT);

                    Attachment attachmentBean = (Attachment) i.next() ;
                    String mimeType = attachmentBean.getMimeType() ;
                    Object data = null ;

                    /*if (attachmentBean.getAttachment() instanceof URL) {
                        logger.debug("Attachment is a URL.") ;
                        URL attachmentUrl = (URL)
                                attachmentBean.getAttachment() ;

                        // Construct a client to make the request
                        HttpClient client = new HttpClient() ;
                        client.startSession(attachmentUrl) ;

                        // make the request and get the response
                        int queryIdx = attachmentUrl.getFile().indexOf("?") ;
                        GetMethod get = null ;
                        if (queryIdx < 0) {
                            get = new GetMethod(attachmentUrl.getFile()
                                    .substring(0, queryIdx)) ;
                        } else {
                            get = new GetMethod(attachmentUrl.getFile()) ;
                        }
                        logger.debug("Get method path=" + get.getPath());
                        if (queryIdx > 0) {
                            get.setQueryString(attachmentUrl.getFile()
                                    .substring(attachmentUrl.getFile()
                                    .indexOf("?") + 1)) ;
                        }
                        get.setStrictMode(false) ; // TODO: parameterise
                        // already done in GetMethod
                        get.setFollowRedirects(true) ;
                        int httpCode = client.executeMethod(get) ;
                        logger.info("Response code was: " + httpCode);

                        data = get.getResponseBodyAsString() ;
                        logger.debug("Response was: " + data);
*/
                        /*
                         * Some headers may be multi-valued,
                         * but content type is not one of them
                         */
  /*                      Header contentType =
                                get.getResponseHeader("Content-Type") ;
                        if (contentType != null
                                && contentType.getValues() != null) {
                            logger.info("Response content header: "
                                    + contentType.getValues()[0].toString());
                            mimeType = contentType.getValues()[0].toString() ;
                        } else {
                            logger.error("URL response did not "
                                    + " include content type") ;
                            // TODO: Often get no content type header
                        }
                    } else {
                        logger.debug("Attachment is a:" + attachmentBean
                                .getAttachment().getClass().getName()) ;
                        data = (String) attachmentBean.getAttachment();
                    }
*/
                    Object obj = attachmentBean.getAttachment() ;
                    if (obj instanceof URL) {
                        logger.debug("Attachment is URL") ;
                        URLDataSource ds = new URLDataSource((URL) obj);
                        DataHandler dh = new DataHandler(ds) ;
                        data = dh.getContent() ;
                        mimeType = dh.getContentType() ;
                        String fileName = ((URL) obj).getFile() ;
                        if (fileName.startsWith("/")) {
                            fileName = fileName.substring(1) ;
                        }
                        attachment.setFileName(fileName) ;
                        attachment.setDataHandler(dh);
                        attachment.setDescription(fileName);
                        String type = dh.getContentType();
                        logger.debug("detected content type: " + type) ;
                        if (type.startsWith("text")) {
                            logger.debug("attaching as attachment") ;
                            attachment.setDisposition(Part.ATTACHMENT);
                        } else {
                            logger.debug("attaching inline") ;
                            attachment.setDisposition(Part.INLINE);
                        }

                        attachment.setHeader("Content-Type", type);
                    } else if (obj instanceof File) {
                        logger.debug("Attachment is File") ;
                        File file = (File) obj ;
                        FileDataSource ds = new FileDataSource(file);
                        DataHandler dh = new DataHandler(ds) ;
                        data = dh.getContent() ;
                        mimeType = dh.getContentType() ;
                        attachment.setDescription(file.getName());
                        attachment.setFileName(file.getName()) ;
                        attachment.setDataHandler(dh);
                        String type = dh.getContentType();
                        logger.debug("detected content type: " + type) ;
                        if (type.startsWith("text")) {
                            logger.debug("attaching as attchment") ;
                            attachment.setDisposition(Part.ATTACHMENT);
                        } else {
                            logger.debug("attaching inline") ;
                            attachment.setDisposition(Part.INLINE);
                        }

                        attachment.setHeader("Content-Type", type);
                    } else if (obj instanceof String) {
                        logger.debug("Attachment is String") ;
                        data = obj ;
                        mimeType = MT_PLAIN_TEXT ;
                    } else if (obj instanceof Object) {
                        logger.debug("Attachment is Object") ;
                        DataHandler dh = new DataHandler(obj, MT_BINARY) ;
                        data = dh.getContent() ;
                        attachment.setFileName(attachmentBean.getName()) ;
                        attachment.setDataHandler(dh);
                        attachment.setDescription(attachmentBean.getName());
                        logger.debug("attaching inline") ;
                        attachment.setDisposition(Part.INLINE);
                        mimeType = MT_BINARY ;
                        attachment.setHeader("Content-Type", mimeType);
                    } else {
                        data = "Unsupported attachment type: "
                                + obj.getClass().getName() ;
                        mimeType = MT_PLAIN_TEXT ;
                        logger.error("Unsupported attachment"
                                + obj.getClass().getName()) ;
                    }
                    attachment.setContent(data, mimeType) ;
                    holder.addBodyPart(attachment);
                    // TODO Having this here means it is done repeatedly,
                    // should do no harm <shrug>
                    holder.setSubType("mixed");
                }
            } catch (org.apache.commons.httpclient.HttpException e) {
                logger.error("Error handling request: " + e.getMessage());
            } catch (IOException ioe) {
                logger.error(ioe.toString());
            }

            // add to base message and send
            msg.setContent(holder);
            Transport.send(msg);

        } catch (MessagingException e) {
            logger.error("MessagingException: " + e.getMessage(), e) ;
        }

        logger.info("METHOD_EXIT sendEmail");
    }

    /**
     * Constructs a <code>Multipart</code> containing the essential basic info
     *
     * @param to Target email address.
     * @param from Message 'from' address. If supplied this overrides the
     *      default in the JNDI context.
     * @param subject Subject for the mail message.
     * @param message Body text for the mail message.
     * @return Message.
     */
    private Message getBaseMessage (String to, String from,
            String subject, String message) {
        logger.info("METHOD_ENTRY getBaseMessage");

        if (from == null) {
            from = this.from ;
        }

        //URLName url   = new URLName(protocol, host, -1, mbox, user, passwd);
        Properties props = new Properties();
        // XXX - could use Session.getTransport() and Transport.connect()
        // XXX - assume we're using SMTP
        //props.put("mail.host", host);
        props.put("mail.user", user);
        props.put("mail.smtp.host", host);

        // Get a Session object
        session = Session.getDefaultInstance(props, null);
        if (session == null) {
            logger.error("Mail session instantiation failed with properties:");
            logger.error("props=" + props.toString()) ;
        } else {
            logger.info("Mail session instantiation succeeded");
            logger.info("props=" + props.toString()) ;
        }
        session.setDebug(debug.booleanValue());

        // Construct the message
        Message msg = new MimeMessage(session);
        try {
            msg.setFrom(new InternetAddress(from));
            msg.setRecipients(Message.RecipientType.TO,
                    InternetAddress.parse(to, false));
//          msg.setRecipients(Message.RecipientType.CC,
//                  InternetAddress.parse(cc, false));
//          msg.setRecipients(Message.RecipientType.BCC,
//                  InternetAddress.parse(bcc, false));
            msg.setSubject(subject == null ? blankSubject : subject);
            msg.setContent(blankMessage == null ? blankMessage : message,
                    MT_PLAIN_TEXT);
            msg.setSentDate(new Date());
            msg.setHeader("X-Mailer", mailer);
        } catch (AddressException ae) {
            logger.error("AddressException: " + ae.getMessage()) ;
            logger.error(ae) ;
        } catch (MessagingException me) {
            logger.error("MessagingException: " + me.getMessage()) ;
            logger.error(me) ;
        }

        logger.info("METHOD_EXIT getBaseMessage");
        return msg ;
    }

    /**
     * Converts a URL reference to an InputStream
     *
     * <p>The URL must be accessable using the http protocol. </p>
     *
     * <p>Makes use of the commons-httpclient library
     * <a href="http://jakarta.apache.org/commons/httpclient">
     * commons-httpclient library homepage</a>.</p>
     * 
     * <p>This method is not used as the URL datasource in the activation 
     * library seems to do an adequate job. It has not yet been tested on 
     * https URLs; if it works there too this will be removed.</p>
     * 
     * @param url The URL to fetch
     * @return The content of the URL. 
     */
    private Object getUrlContent(URL url) {
        logger.info("METHOD_ENTRY getUrlContent, url=" + url.toExternalForm());

        String data = new String();
        try {
            // Construct a client to make the request
            HttpClient client = new HttpClient() ;
            client.startSession(url) ;

            // make the request and get the response
            GetMethod get = new GetMethod(url.toExternalForm()) ;
            int httpCode = client.executeMethod(get) ;

            switch (httpCode) {
            case 100:
            /*
             * According to the W3C this is ok,
             * response will be received in due course:
             * http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.1.1
             */
                logger.debug("Received HTTP code 100");
                break ;

            case 200:
                logger.debug("Received HTTP code 200");
                break ;
            default :
                logger.error("HTTP code was " + httpCode) ;
            }


            data = get.getResponseBodyAsString() ;
            logger.debug("Response was: " + data);

        } catch (org.apache.commons.httpclient.HttpException e) {
            logger.error("Error handling request: " + e.getMessage());
        } catch (IOException ioe) {
            logger.error(ioe.toString());
        }

        logger.info("METHOD_EXIT: getUrlContent");
        return data ;
    }

// Members  -------------------------------------------------------------------

    /**
     * The Log4J <code>Logger</code> doing the logging.
     */
    private static Logger logger = Logger.getLogger(EmailBean.class);

    /** The Mail protocol */
    private String protocol ;

    /** Mailbox to read from (not used) */
    private String mbox     ;

    /** JavaMail property 'send' */
    private String send     ;

    /** The Mail host */
    private String host     ;

    /** The mail user to authenticate as */
    private String user     ;

    /** Password for the Mail user */
    private String passwd   ;

    /** JavaMail property 'mailer' */
    private String mailer   ;

    /** Default 'from' address, if not supplied by caller */
    private String from     ;

    /** Whether to log debug messages from the Mail session */
    private Boolean debug = Boolean.TRUE    ;

    /**
     * String to use when caller does not specify a Subject.
     */
    private String blankSubject = "[No Subject]" ;

    /**
     * String to use when caller does not specify a Message text.
     */
    private String blankMessage = "[No Message]" ;

    /**
     * Whether bean has already been initialised.
     */
    private boolean initialised = false ;

    /**
     *
     */
    private Session session ;

// Constant member properties -------------------------------------------------

    /** A valid message content type */
    public static final String MT_PLAIN_TEXT = "text/plain";

    /** A valid message content type */
    public static final String MT_HTML = "text/html";

    /** A valid message content type */
    public static final String MT_BINARY = "application/octet-stream" ;

    /** CVS info about this class and its current version */
    public static final String ABOUT =
            "$Id: EmailBean.java,v 1.12 2003/01/27 08:00:53 TimAndVicky Exp $ $Revision: 1.12 $" ;

}
