/* 
 *     PROJECT          : emailEJB 
 *
 *     COPYRIGHT        : Copyright (C) 1999-2003 Tim Stephenson
 *
 *     contact me at gpl@enableit.org.
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 
package org.enableit.tools.email.handlers;

// Java Imports
import java.awt.datatransfer.DataFlavor ;
import java.io.BufferedOutputStream ;
import java.io.IOException ;
import java.io.InputStream ;
import java.io.OutputStream ;
import javax.activation.ActivationDataFlavor ;
import javax.activation.DataContentHandler ;
import javax.activation.DataSource ;

// Log4J Imports
import org.apache.log4j.Category;

/**
 * Data content handler for binary data types.
 *
 * <p>The J2EE documentation says that 'application/octet-stream' should
 * be specified as the mime type for unknown data types. This DCH, once
 * registered, while handle that type.</p>
 * 
 * @author Tim Stephenson 
 */
public class OctetStreamDataContentHandler implements DataContentHandler {

/*
 * Properties
 */
    /**
     * The Log4J <code>Category</code> doing the logging.
     */
    private static Category logger =
            Category.getInstance(OctetStreamDataContentHandler.class);

    /**
     * Internal ref to the DataFlavor I am handler for.
     */
    private static ActivationDataFlavor myDF;

/*
 * Constructors
 */
    /**
     * Default Constructor.
     */
    public OctetStreamDataContentHandler() {

    }

/*
 * Methods
 */

    /**
     * Return an object representing the data in its most preferred form.
     * @param ds The datasource to get the content of. 
     * @return The content (as a <code>byte[]</code>)
     * @throws IOException If there is a problem in reading the input.
     */
    public Object getContent(DataSource ds)
            throws IOException {
        logger.info("METHOD_ENTRY: getContent");

        InputStream is = ds.getInputStream() ;
        logger.debug("InputStream is of type" + is.getClass().getName());
        int j;
        byte data[] = new byte[1024];
        byte dataCopy[];
        // read 1Kb at a time into byte array
        for (int i = 0; (j = is.read(data, i, 1024)) != -1; data = dataCopy) {
            i += j;
            dataCopy = new byte[i + 1024];
            System.arraycopy(data, 0, dataCopy, 0, i);
        }

        logger.info("METHOD_EXIT: getContent");
        return data;
    }

    /**
     * Returns an object which represents the data to be transferred.
     * @param df The <code>DataFlavor</code> the datasource represents. 
     * @param ds The datasource to get the content of. 
     * @return The content (as a <code>byte[]</code>) or null if the 
     *      DataFlavor does not match my internal one.
     * @throws IOException If there is a problem in reading the input.
     */
    public Object getTransferData(DataFlavor df, DataSource ds)
            throws IOException {
        logger.info("METHOD_ENTRY: getTransferData");
        Object obj = null ;
        if (myDF.equals(df)) { 
            obj = getContent(ds);
        } else {
            obj = null;
        } 
        logger.info("METHOD_EXIT: getTransferData");
        return obj ;
    }

    /**
     * Returns an array of DataFlavor objects indicating the flavors the data
     * can be provided in.
     * @return The <code>DataFlavors</code> this 
     *      <code>DataContentHandler</code> can handle. 
     */
    public DataFlavor[] getTransferDataFlavors() {
        logger.info("METHOD_ENTRY: getTransferDataFlavors");
        DataFlavor[] df = new DataFlavor[] { myDF };
        logger.info("METHOD_EXIT: getTransferDataFlavors");
        return df ;
    }

    /**
     * Convert the object to a byte stream of the specified MIME type and
     * write it to the output stream.
     * @param obj The content object (as a <code>byte[]</code>) or null if the 
     *      DataFlavor does not match my internal one.
     * @param mimeType The mime type of the data object.
     * @param os <code>OutputStream</code> to write the data object to.
     * @throws IOException If there is a problem in writing the output.
     */
    public void writeTo(Object obj, String mimeType, OutputStream os)
            throws IOException {
        logger.info("METHOD_ENTRY: writeTo");
        if (!(obj instanceof byte[])) { 
            throw new IOException("\"" + myDF.getMimeType()
                    + "\" DataContentHandler requires byte[] object, "
                    + "was given object of type "
                    + obj.getClass().toString());
        } 
        byte[] data = (byte[]) obj ;
        BufferedOutputStream out = new BufferedOutputStream(os) ;
        out.write(data) ;
        out.flush();
        logger.info("METHOD_EXIT: writeTo");
    }

}
